import csv
import json
import sqlite3
import hashlib
import asyncio
from dataclasses import dataclass
from structlog import get_logger

log = get_logger()

"""
Implementation thoughts

Originally beaker was the central object, but pretty quickly
most functionality was moved to recipe. Now beaker is just a
thin wrapper around a sqlite table.

Recipe is the central object. It contains a list of pours,
which are the instructions for how to transform data from one
beaker to another.

Recipe also contains a list of beakers, which are just sqlite
tables. Beakers are created by the recipe, and are passed to
pours as arguments.

Beakers right now know they are stored in SQLite, but I think
that will be abstracted away. Beakers should be able to be
stored in any database, or on disk.

Core functionality of a beaker:
- initialize (create table if it doesn't exist)
- add item with optional link to another item
    - should this be links? (probably)
- iterate over items
- get item by id
- count items

Ultimately, this is a graph problem. Beakers are nodes, and
pours are edges. The recipe is the graph.
"""


class Beaker:
    def __init__(self, table_name: str, recipe):
        self.table_name = table_name
        self.recipe = recipe

        # create table if it doesn't exist
        self.recipe.cursor.execute(
            f"CREATE TABLE IF NOT EXISTS {self.table_name} (id INTEGER PRIMARY KEY, data JSON, from_table TEXT NULL, from_id INTEGER NULL)"
        )

    def __repr__(self):
        return f"Beaker({self.table_name})"

    def items(self):
        self.recipe.cursor.execute(f"SELECT id, data FROM {self.table_name}")
        data = self.recipe.cursor.fetchall()
        for item in data:
            yield item["id"], json.loads(item["data"])

    def __len__(self):
        self.recipe.cursor.execute(f"SELECT COUNT(*) FROM {self.table_name}")
        return self.recipe.cursor.fetchone()[0]

    def add_item(self, item: dict, from_table=None, from_id=None) -> None:
        self.recipe.cursor.execute(
            f"INSERT INTO {self.table_name} (data) VALUES (?)", (json.dumps(item),)
        )
        self.recipe.cursor.commit()


def get_sha512(filename: str) -> str:
    with open(filename, "rb") as file:
        return hashlib.sha512(file.read()).hexdigest()


@dataclass
class Pour:
    from_beaker: str
    to_beaker: str
    transform: callable


@dataclass
class Split:
    from_beaker: str
    condition: callable
    if_true: str
    if_false: str


class Recipe:
    def __init__(self, name: str):
        self.name = name
        self.beakers = {}
        self.pours = []
        self.splits = []
        self.db = sqlite3.connect("beakers.db")
        self.cursor = self.db.cursor()
        self.cursor.row_factory = sqlite3.Row
        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS _metadata (table_name TEXT PRIMARY KEY, data JSON)"
        )

    def __repr__(self) -> str:
        return f"Recipe({self.name})"

    def get_metadata(self, table_name) -> dict:
        self.cursor.execute(
            "SELECT data FROM _metadata WHERE table_name = ?",
            (table_name,),
        )
        try:
            data = self.cursor.fetchone()["data"]
            log.debug("get_metadata", table_name=table_name, data=data)
            return json.loads(data)
        except TypeError:
            log.debug("get_metadata", table_name=table_name, data={})
            return {}

    def save_metadata(self, table_name: str, data: dict) -> None:
        data_json = json.dumps(data)
        log.info("save_metadata", table_name=table_name, data=data_json)
        # sqlite upsert
        self.cursor.execute(
            "INSERT INTO _metadata (table_name, data) VALUES (?, ?) ON CONFLICT(table_name) DO UPDATE SET data = ?",
            (table_name, data_json, data_json),
        )
        self.db.commit()

    def add_pour(self, from_beaker: Beaker, to_beaker: Beaker, transform: callable):
        pour = Pour(from_beaker, to_beaker, transform)
        self.pours.append(pour)

    def csv_to_beaker(self, filename: str, beaker: Beaker) -> None:
        lg = log.bind(beaker=beaker, filename=filename)
        # three cases: empty, match, mismatch
        # case 1: empty
        if len(beaker) == 0:
            with open(filename, "r") as file:
                reader = csv.DictReader(file)
                added = 0
                for row in reader:
                    beaker.add_item(row)
                    added += 1
            lg.info("from_csv", case="empty", added=added)
            meta = self.get_metadata(beaker.table_name)
            meta["sha512"] = get_sha512(filename)
            self.save_metadata(beaker.table_name, meta)
        else:
            old_sha = self.get_metadata(beaker).get("sha512")
            new_sha = get_sha512(filename)
            if old_sha != new_sha:
                # case 3: mismatch
                lg.info("from_csv", case="mismatch", old_sha=old_sha, new_sha=new_sha)
                raise Exception("sha512 mismatch")
            else:
                # case 2: match
                lg.info("from_csv", case="match")
        return beaker

    def declare_beaker(self, name: str, temp=True):
        beaker = Beaker(name, self)
        self.beakers[name] = beaker
        if temp:
            self.cursor.execute(f"DROP TABLE IF EXISTS {name}")
        return beaker

    def add_pour(self, from_beaker: str, to_beaker: str, transform: callable):
        pour = Pour(from_beaker, to_beaker, transform)
        self.pours.append(pour)

    def add_split(
        self, from_beaker: str, condition: callable, if_true: str, if_false: str
    ):
        split = Split(from_beaker, condition, if_true, if_false)
        self.splits.append(split)
        return split

    def run_linearly(self):
        log.info("recipe", recipe=self)
        loop = asyncio.get_event_loop()

        for pour in self.pours:
            log.info(
                "pour",
                from_beaker=pour.from_beaker,
                to_beaker=pour.to_beaker,
                to_pour=len(pour.from_beaker),
            )
            for id, item in pour.from_beaker.items():
                log.info("pour_item", id=id, item=item)
                transformed = loop.run_until_complete(pour.transform(item))
                pour.to_beaker.add_item(transformed, pour.from_beaker.table_name, id)
